@@ -29,6 +29,7 @@
 #include <unistd.h>
 #endif // HAVE_UNISTD_H
 #include <sys/resource.h>
+#include <sys/wait.h>
 #include <grp.h>
 
 #include <cinttypes>
@@ -38,6 +39,8 @@
 
 #include <ev.h>
 
+#include <ares.h>
+
 #include "shrpx_config.h"
 #include "shrpx_connection_handler.h"
 #include "shrpx_log_config.h"
@@ -48,10 +51,12 @@
 #include "shrpx_memcached_dispatcher.h"
 #include "shrpx_memcached_request.h"
 #include "shrpx_process.h"
-#include "shrpx_ssl.h"
+#include "shrpx_tls.h"
+#include "shrpx_log.h"
 #include "util.h"
 #include "app_helper.h"
 #include "template.h"
+#include "xsi_strerror.h"
 
 using namespace nghttp2;
 
@@ -62,33 +67,39 @@ void drop_privileges(
 #ifdef HAVE_NEVERBLEED
     neverbleed_t *nb
 #endif // HAVE_NEVERBLEED
-    ) {
-  if (getuid() == 0 && get_config()->uid != 0) {
-    if (initgroups(get_config()->user.c_str(), get_config()->gid) != 0) {
+) {
+  std::array<char, STRERROR_BUFSIZE> errbuf;
+  auto config = get_config();
+
+  if (getuid() == 0 && config->uid != 0) {
+#ifdef HAVE_NEVERBLEED
+    if (nb) {
+      neverbleed_setuidgid(nb, config->user.c_str(), 1);
+    }
+#endif // HAVE_NEVERBLEED
+
+    if (initgroups(config->user.c_str(), config->gid) != 0) {
       auto error = errno;
       LOG(FATAL) << "Could not change supplementary groups: "
-                 << strerror(error);
+                 << xsi_strerror(error, errbuf.data(), errbuf.size());
       exit(EXIT_FAILURE);
     }
-    if (setgid(get_config()->gid) != 0) {
+    if (setgid(config->gid) != 0) {
       auto error = errno;
-      LOG(FATAL) << "Could not change gid: " << strerror(error);
+      LOG(FATAL) << "Could not change gid: "
+                 << xsi_strerror(error, errbuf.data(), errbuf.size());
       exit(EXIT_FAILURE);
     }
-    if (setuid(get_config()->uid) != 0) {
+    if (setuid(config->uid) != 0) {
       auto error = errno;
-      LOG(FATAL) << "Could not change uid: " << strerror(error);
+      LOG(FATAL) << "Could not change uid: "
+                 << xsi_strerror(error, errbuf.data(), errbuf.size());
       exit(EXIT_FAILURE);
     }
     if (setuid(0) != -1) {
       LOG(FATAL) << "Still have root privileges?";
       exit(EXIT_FAILURE);
     }
-#ifdef HAVE_NEVERBLEED
-    if (nb) {
-      neverbleed_setuidgid(nb, get_config()->user.c_str(), 1);
-    }
-#endif // HAVE_NEVERBLEED
   }
 }
 } // namespace
@@ -105,16 +116,16 @@ void graceful_shutdown(ConnectionHandler *conn_handler) {
 
   conn_handler->disable_acceptor();
 
-  // After disabling accepting new connection, disptach incoming
+  // After disabling accepting new connection, dispatch incoming
   // connection in backlog.
 
   conn_handler->accept_pending_connection();
 
   conn_handler->graceful_shutdown_worker();
 
-  if (get_config()->num_worker == 1) {
-    if (conn_handler->get_single_worker()->get_worker_stat()->num_connections ==
-        0) {
+  auto single_worker = conn_handler->get_single_worker();
+  if (single_worker) {
+    if (single_worker->get_worker_stat()->num_connections == 0) {
       ev_break(conn_handler->get_loop());
     }
 
@@ -127,12 +138,13 @@ namespace {
 void reopen_log(ConnectionHandler *conn_handler) {
   LOG(NOTICE) << "Reopening log files: worker process (thread main)";
 
-  (void)reopen_log_files();
-  redirect_stderr_to_errorlog();
+  auto config = get_config();
+  auto &loggingconf = config->logging;
 
-  if (get_config()->num_worker > 1) {
-    conn_handler->worker_reopen_log_files();
-  }
+  (void)reopen_log_files(loggingconf);
+  redirect_stderr_to_errorlog(loggingconf);
+
+  conn_handler->worker_reopen_log_files();
 }
 } // namespace
 
@@ -152,7 +164,7 @@ void ipc_readcb(struct ev_loop *loop, ev_io *w, int revents) {
   if (nread == 0) {
     // IPC socket closed.  Perform immediate shutdown.
     LOG(FATAL) << "IPC socket is closed.  Perform immediate shutdown.";
-    _Exit(EXIT_FAILURE);
+    nghttp2_Exit(EXIT_FAILURE);
   }
 
   for (ssize_t i = 0; i < nread; ++i) {
@@ -217,7 +229,8 @@ void renew_ticket_key_cb(struct ev_loop *loop, ev_timer *w, int revents) {
 
     auto max_tickets =
         static_cast<size_t>(std::chrono::duration_cast<std::chrono::hours>(
-                                get_config()->tls.session_timeout).count());
+                                get_config()->tls.session_timeout)
+                                .count());
 
     new_keys.resize(std::min(max_tickets, old_keys.size() + 1));
     std::copy_n(std::begin(old_keys), new_keys.size() - 1,
@@ -262,8 +275,7 @@ void memcached_get_ticket_key_cb(struct ev_loop *loop, ev_timer *w,
   auto req = make_unique<MemcachedRequest>();
   req->key = "nghttpx:tls-ticket-key";
   req->op = MEMCACHED_OP_GET;
-  req->cb = [conn_handler, dispatcher, w](MemcachedRequest *req,
-                                          MemcachedResult res) {
+  req->cb = [conn_handler, w](MemcachedRequest *req, MemcachedResult res) {
     switch (res.status_code) {
     case MEMCACHED_ERR_NO_ERROR:
       break;
@@ -341,13 +353,13 @@ void memcached_get_ticket_key_cb(struct ev_loop *loop, ev_timer *w,
       key.hmac = EVP_sha256();
       key.hmac_keylen = hmac_keylen;
 
-      std::copy_n(p, key.data.name.size(), key.data.name.data());
+      std::copy_n(p, key.data.name.size(), std::begin(key.data.name));
       p += key.data.name.size();
 
-      std::copy_n(p, enc_keylen, key.data.enc_key.data());
+      std::copy_n(p, enc_keylen, std::begin(key.data.enc_key));
       p += enc_keylen;
 
-      std::copy_n(p, hmac_keylen, key.data.hmac_key.data());
+      std::copy_n(p, hmac_keylen, std::begin(key.data.hmac_key));
       p += hmac_keylen;
 
       ticket_keys->keys.push_back(std::move(key));
@@ -374,33 +386,45 @@ void nb_child_cb(struct ev_loop *loop, ev_child *w, int revents) {
 
   LOG(FATAL) << "neverbleed process exitted; aborting now";
 
-  _Exit(EXIT_FAILURE);
+  nghttp2_Exit(EXIT_FAILURE);
 }
 } // namespace
 #endif // HAVE_NEVERBLEED
 
 int worker_process_event_loop(WorkerProcessConfig *wpconf) {
-  if (reopen_log_files() != 0) {
+  int rv;
+  std::array<char, STRERROR_BUFSIZE> errbuf;
+  (void)errbuf;
+
+  auto config = get_config();
+
+  if (reopen_log_files(config->logging) != 0) {
     LOG(FATAL) << "Failed to open log file";
     return -1;
   }
 
+  rv = ares_library_init(ARES_LIB_INIT_ALL);
+  if (rv != 0) {
+    LOG(FATAL) << "ares_library_init failed: " << ares_strerror(rv);
+    return -1;
+  }
+
   auto loop = EV_DEFAULT;
 
-  ConnectionHandler conn_handler(loop);
+  auto gen = util::make_mt19937();
 
-  for (auto &addr : get_config()->conn.listener.addrs) {
+  ConnectionHandler conn_handler(loop, gen);
+
+  for (auto &addr : config->conn.listener.addrs) {
     conn_handler.add_acceptor(make_unique<AcceptHandler>(&addr, &conn_handler));
   }
 
-  auto &upstreamconf = get_config()->conn.upstream;
-
 #ifdef HAVE_NEVERBLEED
-  if (!upstreamconf.no_tls || ssl::downstream_tls_enabled()) {
-    std::array<char, NEVERBLEED_ERRBUF_SIZE> errbuf;
+  {
+    std::array<char, NEVERBLEED_ERRBUF_SIZE> nb_errbuf;
     auto nb = make_unique<neverbleed_t>();
-    if (neverbleed_init(nb.get(), errbuf.data()) != 0) {
-      LOG(FATAL) << "neverbleed_init failed: " << errbuf.data();
+    if (neverbleed_init(nb.get(), nb_errbuf.data()) != 0) {
+      LOG(FATAL) << "neverbleed_init failed: " << nb_errbuf.data();
       return -1;
     }
 
@@ -423,8 +447,8 @@ int worker_process_event_loop(WorkerProcessConfig *wpconf) {
   MemchunkPool mcpool;
 
   ev_timer renew_ticket_key_timer;
-  if (!upstreamconf.no_tls) {
-    auto &ticketconf = get_config()->tls.ticket;
+  if (tls::upstream_tls_enabled(config->conn)) {
+    auto &ticketconf = config->tls.ticket;
     auto &memcachedconf = ticketconf.memcached;
 
     if (!memcachedconf.host.empty()) {
@@ -437,7 +461,7 @@ int worker_process_event_loop(WorkerProcessConfig *wpconf) {
       conn_handler.set_tls_ticket_key_memcached_dispatcher(
           make_unique<MemcachedDispatcher>(
               &ticketconf.memcached.addr, loop, ssl_ctx,
-              StringRef{memcachedconf.host}, &mcpool));
+              StringRef{memcachedconf.host}, &mcpool, gen));
 
       ev_timer_init(&renew_ticket_key_timer, memcached_get_ticket_key_cb, 0.,
                     0.);
@@ -477,9 +501,7 @@ int worker_process_event_loop(WorkerProcessConfig *wpconf) {
     }
   }
 
-  int rv;
-
-  if (get_config()->num_worker == 1) {
+  if (config->single_thread) {
     rv = conn_handler.create_single_worker();
     if (rv != 0) {
       return -1;
@@ -492,12 +514,13 @@ int worker_process_event_loop(WorkerProcessConfig *wpconf) {
 
     rv = pthread_sigmask(SIG_BLOCK, &set, nullptr);
     if (rv != 0) {
-      LOG(ERROR) << "Blocking SIGCHLD failed: " << strerror(rv);
+      LOG(ERROR) << "Blocking SIGCHLD failed: "
+                 << xsi_strerror(rv, errbuf.data(), errbuf.size());
       return -1;
     }
 #endif // !NOTHREADS
 
-    rv = conn_handler.create_worker_thread(get_config()->num_worker);
+    rv = conn_handler.create_worker_thread(config->num_worker);
     if (rv != 0) {
       return -1;
     }
@@ -505,7 +528,8 @@ int worker_process_event_loop(WorkerProcessConfig *wpconf) {
 #ifndef NOTHREADS
     rv = pthread_sigmask(SIG_UNBLOCK, &set, nullptr);
     if (rv != 0) {
-      LOG(ERROR) << "Unblocking SIGCHLD failed: " << strerror(rv);
+      LOG(ERROR) << "Unblocking SIGCHLD failed: "
+                 << xsi_strerror(rv, errbuf.data(), errbuf.size());
       return -1;
     }
 #endif // !NOTHREADS
@@ -515,14 +539,19 @@ int worker_process_event_loop(WorkerProcessConfig *wpconf) {
 #ifdef HAVE_NEVERBLEED
       nb
 #endif // HAVE_NEVERBLEED
-      );
+  );
 
   ev_io ipcev;
   ev_io_init(&ipcev, ipc_readcb, wpconf->ipc_fd, EV_READ);
   ipcev.data = &conn_handler;
   ev_io_start(loop, &ipcev);
 
-  if (!upstreamconf.no_tls && !get_config()->tls.ocsp.disabled) {
+  if (tls::upstream_tls_enabled(config->conn) && !config->tls.ocsp.disabled) {
+    if (config->tls.ocsp.startup) {
+      conn_handler.set_enable_acceptor_on_ocsp_completion(true);
+      conn_handler.disable_acceptor();
+    }
+
     conn_handler.proceed_next_cert_ocsp();
   }
 
@@ -535,11 +564,29 @@ int worker_process_event_loop(WorkerProcessConfig *wpconf) {
   conn_handler.cancel_ocsp_update();
 
 #ifdef HAVE_NEVERBLEED
-  if (nb && nb->daemon_pid != -1) {
-    kill(nb->daemon_pid, SIGTERM);
+  if (nb) {
+    assert(nb->daemon_pid > 0);
+
+    rv = kill(nb->daemon_pid, SIGTERM);
+    if (rv != 0) {
+      auto error = errno;
+      LOG(ERROR) << "Could not send signal to neverbleed daemon: errno="
+                 << error;
+    }
+
+    while ((rv = waitpid(nb->daemon_pid, nullptr, 0)) == -1 && errno == EINTR)
+      ;
+    if (rv == -1) {
+      auto error = errno;
+      LOG(ERROR) << "Error occurred while we were waiting for the completion "
+                    "of neverbleed process: errno="
+                 << error;
+    }
   }
 #endif // HAVE_NEVERBLEED
 
+  ares_library_cleanup();
+
   return 0;
 }
 
